#!/usr/bin/perl -w

# Copyright 2016, 2017, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


# Picture of fAltBpred hulls surrounding m.

use 5.006;
use strict;
use warnings;

use FindBin;
use lib $FindBin::Bin;
use MyTerdragon 'xy_to_node_name', '@dir3_to_dx','@dir3_to_dy',
  'xy_rotate_plus120','$alt_hullstouch_segs';
use MyPlanar;
use MyPrintwrap ':all';
$|=1;

# uncomment this to run the ### lines
# use Smart::Comments;


print "  % generated by tools/$FindBin::Script\n";

# arrayref of points [x,y] which are the convex hull limit.
# Points are scaled by *24 so integers.
# Points are for use in X,Y even forms, so curve end is X=2,Y=0.
our $alt_hull24_points = [ [0,  0],  # left-horiz is -2/24 =   - 1/24*2.
                           [36, -12],
                           [48, 0],
                           [12, 12] ];

our $conf = MyTerdragonConf->new
  (name => 'hullsnontouch',
   segs => [
            '-1,-1,1',
            '4,0,2',
            '4,0,1',
            '-1,1,2',

            '1,1,0',
            '1,1,1',
            '0,2,2',

            '-1,-1,0',
            '3,-1,2',
            '2,-2,1',
           ]);

{
  my $str = $conf->print_to_str;
  $str =~ s/^/  %    /mg;
  print $str;
  print "  %\n";
}
{
  my $union = $conf->union($alt_hullstouch_segs)
    ->set_name('hullstouch and hullsnontouch union');
  my $str = $union->print_to_str;
  $str =~ s/^/  %    /mg;
  print $str;
  print "  %\n";
}
$conf->intersection($alt_hullstouch_segs)->num_segs == 0
  or die "oops";

sub points_rotate_plus120 {
  my ($points, $reps) = @_;
  return [ map {[xy_rotate_plus120(@$_,$reps)]} @$points ];
}

# $seg is a segment as a string "x,y,dir".  Must have x,y even.
# Return an arrayref of points like $hull24_points but moved and rotated
# according to $seg location and direction.
# $hull24_points has all coordinates multiplied by 24 so as to be integers.
# The same is in the return here.
sub seg_to_hull24_points {
  my ($seg) = @_;
  my ($x,$y,$dir) = split /,/, $seg;
  require MyPlanar;
  my $points = points_rotate_plus120($alt_hull24_points, $dir);
  return MyPlanar::points_move($points, 24*$x, 24*$y);
}

{
  my $scale = 24*5;
  my $rectangle = [ [-$scale*3,  -$scale*3],
                    [ $scale*4,  -$scale*3],
                    [ $scale*4,   $scale*3],
                    [-$scale*3,   $scale*3]];
  my $inside_gpc = MyPlanar::points_to_gpc($rectangle);
  foreach my $x (-5 .. 6) {
    foreach my $y (-5 .. 5) {
      foreach my $dir (0 .. 2) {
        next unless ($x+$y)%2 == 0;
        my $seg = "$x,$y,$dir";
        next if $alt_hullstouch_segs->contains($seg)
          && $seg ne '0,0,0';
        my $seg_points = seg_to_hull24_points($seg);
        $seg_points = MyPlanar::points_scale($seg_points, $scale/24);
        my $seg_gpc = MyPlanar::points_to_gpc($seg_points);
        $inside_gpc = Math::Geometry::Planar::GpcClip('DIFFERENCE',
                                                      $inside_gpc, $seg_gpc);
      }
    }
  }

  my @inside_points = MyPlanar::gpc_to_points_list($inside_gpc);
  ### @inside_points

  my @points_list;
  my @contours = Math::Geometry::Planar::Gpc2Polygons($inside_gpc);
  # @contours==1 or die "oops, expected one contours, got ",scalar(@contours);
  my $total_holes = 0;
  foreach my $contour (@contours) {  # Math::Geometry::Planar contour
    ### $contour
    # Second and subsequent polygon of the contour are holes.
    # Think their points already cycle the opposite way as required by pgf
    # default \fill[nonzero rule], but using \fill[even odd rule] works either
    # way.
    print "  \\begin{scope}[xscale=1/2/$scale,yscale=sqrt(3)/2/$scale]\n";
    $Printwrap = 2;
    printwrap(" \\fill[my grey,even odd rule]");
    $Printwrap += 2;
    my $polygons = $contour->polygons;  # arrayref of points arrayrefs
    $total_holes += scalar(@$polygons)-1;
    my $separator = '';
    foreach my $points (@$polygons) {
      ### $points
      printwrap($separator);
      $separator = "\n";
      printwrap_poly($points);
      push @points_list, @$points;
    }
    printwrap(";\n");
    print "  \\end{scope}\n";
  }

  my $inside_polygons = $contours[0]->polygons;
  ### inside_polygons: scalar(@$inside_polygons)
  ### $inside_polygons

  my $inside_hole_points = $inside_polygons->[1];
  # $inside_hole_points && scalar(@$inside_hole_points) == 10 or warn "oops";

  # $conf is all hulls which have some boundary touch of the $inside_gpc
  # formed (but are not in $alt_hullstouch_segs);
  #
  my $inside_points = $inside_polygons->[0];
  foreach my $x (-3 .. 4) {
    foreach my $y (-3 .. 3) {
      foreach my $dir (0 .. 2) {
        next unless ($x+$y)%2 == 0;
        my $seg = "$x,$y,$dir";
        next if $alt_hullstouch_segs->contains($seg);
        my $seg_points = seg_to_hull24_points($seg);
        $seg_points = MyPlanar::points_scale($seg_points, $scale/24);

        my $by_segs    = $conf->contains($seg) ? 1 : 0;
        my $by_polygon = MyPlanar::points_has_boundary_touch($inside_points,
                                                             $seg_points)
          ? 1 : 0;
        unless ($by_polygon == $by_segs) {
          # warn "oops, inside_gpc vs conf at $seg";
        }
      }
    }
  }

  # Points scaled to end at 1, and no extra sqrt(3) on Y.
  my $m_points_1 = MyPlanar::points_xyscale($alt_hull24_points,
                                            1/24 / 2,
                                            1/24 * sqrt(3)/2);
  my $inside_points_1 = MyPlanar::points_xyscale($inside_points,
                                                 1/$scale * 1/2,
                                                 1/$scale * sqrt(3)/2);
  print "% GP-DEFINE  hullsnontouch_min_distance = ",
    MyPlanar::poly_distance_boundary_to_boundary($inside_points_1,
                                                 $m_points_1),
                                                   ";\n";
}


my %seen;
foreach my $seg ('0,0,0', sort $conf->segs_list) {
  my ($x,$y,$dir) = split /,/, $seg;
  foreach (0, 1) {
    my $name = xy_to_node_name($x,$y);
    unless ($seen{$name}++) {
      my $blank = ($x==0 && $y==0) || ($x==1 && $y==0) ? '' : ' blank';
      print "  \\node ($name) at ($x/2,{$y*sqrt(3)/2}) [my$blank dot];\n";
    }
    $x += $dir3_to_dx[$dir];
    $y += $dir3_to_dy[$dir];
  }
}

foreach my $seg ('0,0,0', sort $conf->segs_list) {
  my ($x,$y,$dir) = split /,/, $seg;
  my $from = xy_to_node_name($x,$y);
  my $to   = xy_to_node_name($x+$dir3_to_dx[$dir], $y+$dir3_to_dy[$dir]);
  my $angle = 120*$dir;
  my $style = ($seg eq '0,0,0' ? 'my m arrow' : 'my arrow');
  print "  \\draw[$style] ($from)--($to);\n";
  print "  \\pic at ($from) [transform shape,rotate=$angle] {my alt hull limit pic};\n";
}

print "  % end generated\n";
exit 0;
