#!/usr/bin/perl -w

# Copyright 2015, 2016, 2017, 2018 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use FindBin;
use Math::PlanePath::TerdragonCurve 117; # v.117 for level_to_n_range()

BEGIN { unshift @INC, $FindBin::Bin }
use MyTerdragon 'xy_rotate_plus60','xy_rotate_minus60','xy_rotate_minus120';
use MyLineSegments;
use MyPrintwrap ':all';


my $level = 8;

if (@ARGV) { $level = shift @ARGV; }
$level % 2 == 0 or die "level must be even";
print "  % generated by tools/$FindBin::Script\n";
print "  \\newcommand{\\MyLevel}{$level}\n";

my $path = Math::PlanePath::TerdragonCurve->new;
my ($n_lo, $n_hi) = $path->level_to_n_range($level);
my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);

sub xy_is_boundary_point {
  my ($x,$y) = @_;
  my @n_list = $path->xy_to_n_list($x,$y);
  foreach my $n (@n_list) {
    foreach my $offset (0, -1) {
      if ($path->_UNDOCUMENTED__n_segment_is_right_boundary($n+$offset)
          || $path->_UNDOCUMENTED__n_segment_is_left_boundary($n+$offset,
                                                              $level)) {
        return 1;
      }
    }
  }
  return 0;
}

sub transform {
  my ($x,$y) = @_;
  foreach (1 .. $level/2) {
    ($x,$y) = xy_rotate_minus60($x,$y);
  }
  return ($x,$y);
}

my $segments = MyLineSegments->new;
{
  my ($prev_x,$prev_y) = $path->n_to_xy($n_lo);
  foreach my $n ($n_lo+1 .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    $segments->add($prev_x,$prev_y, $x,$y);
    ($prev_x,$prev_y) = ($x,$y);
  }
}

$Printwrap = 2;
printwrap("\\draw[my grey] ");
$Printwrap += 2;
$segments->printwrap (transform => \&transform);
printwrap(";\n");

my $line = MyLineSegments->new;

my @spiral;
my $middle = 0;
if ($level == 6) {
  @spiral = (1, 0, 4, 7, 4);
  $middle = 7;
} elsif ($level == 8) {
  @spiral = (1, 0, 4, 7, 24, 10);
  $middle = 21;
} elsif ($level == 10) {
  @spiral = (1, 0, 4, 7, 25, 70, 31);
  $middle = 61;
}

my $EndLength = 0;
my $dx = 2;
my $dy = 0;
my $x = 0;
my $y = 0;
foreach my $len (@spiral) {
  my $boundary;
  foreach (1 .. $len) {
    my $x2 = $x + $dx;
    my $y2 = $y + $dy;
    if (! $segments->is_traversed($x,$y, $x2,$y2)) {
      die "no segment at $x,$y to $x2,$y2";
    }
    if (! $segments->is_traversed($x_hi-$x,$y_hi-$y, $x_hi-$x2,$y_hi-$y2)) {
      die;
    }
    $line->add($x,$y, $x2,$y2);
    $line->add($x_hi-$x,$y_hi-$y, $x_hi-$x2,$y_hi-$y2);
    $EndLength += 2;
    $boundary ||= xy_is_boundary_point($x,$y);

    ($x,$y) = ($x2,$y2);
  }
  if ($len && ! $boundary) {
    die "spiral length $len not on boundary";
  }
  ($dx,$dy) = xy_rotate_plus60($dx,$dy);
}

($dx,$dy) = xy_rotate_minus120($dx,$dy);
foreach (1 .. $middle) {
  my $x2 = $x + $dx;
  my $y2 = $y + $dy;
  if (! $segments->is_traversed($x,$y, $x2,$y2)) {
    die "no middle segment $x,$y to $x2,$y2";
  }
  $line->add($x,$y, $x2,$y2);
  $EndLength++;
  ($x,$y) = ($x2,$y2);
}

$Printwrap = 2;
printwrap("\\draw[my line] ");
$Printwrap += 2;
$line->printwrap (transform => \&transform);
printwrap(";\n");
print "% GP-Test  EndLength($level)==$EndLength\n";


{
  my ($x,$y) = transform(0,0);
  print "  \\coordinate (start) at ($x,$y);\n";
}
{
  my ($x,$y) = transform($x_hi,$y_hi);
  print "  \\coordinate (end) at ($x,$y);\n";
}


print "  \\newcommand{\\MyEndLength}{$EndLength}\n";
print "  % end generated\n";
exit 0;
